# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "asdf-pydantic"
copyright = "2023, Keto D. Zhang"
author = "Keto D. Zhang"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["myst_parser", "autoapi.extension"]
autoapi_dirs = ["../asdf_pydantic"]
myst_enable_extensions = ["tasklist"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_static_path = ["_static"]



from datetime import datetime
from typing import Any, Optional

import asdf
import pytest
from asdf.extension import Extension
from astropy.time import Time

from asdf_pydantic import AsdfPydanticConverter, AsdfPydanticModel


class UnionObject(AsdfPydanticModel):
    _tag = "asdf://asdf-pydantic/examples/tags/unionobject-1.0.0"

    # Order of type matters
    int_or_str: int | str = 0
    datetime_or_time: datetime | Time = datetime(2023, 1, 1)
    collection: tuple | list = tuple()
    anything: Optional[Any] = None


def setup_module():
    AsdfPydanticConverter.add_models(UnionObject)

    class TestExtension(Extension):
        extension_uri = "asdf://asdf-pydantic/examples/extensions/test-1.0.0"

        converters = [AsdfPydanticConverter()]  # type: ignore
        tags = [*AsdfPydanticConverter().tags]  # type: ignore

    asdf.get_config().add_extension(TestExtension())


def test_str_converts_to_str(tmp_path):
    af = asdf.AsdfFile({"obj": UnionObject(int_or_str="test")})
    af.write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af["obj"].int_or_str, str)


def test_integer_converts_to_int(tmp_path):
    asdf.AsdfFile({"obj": UnionObject(int_or_str=0)}).write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af["obj"].int_or_str, int)


def test_datetime_converts_to_datetime(tmp_path):
    af = asdf.AsdfFile({"obj": UnionObject(datetime_or_time=datetime(2023, 1, 1, 0))})
    af.write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af["obj"].datetime_or_time, datetime)


def test_Time_converts_to_Time(tmp_path):
    af = asdf.AsdfFile(
        {
            "obj": UnionObject(
                datetime_or_time=Time("2023-01-01T01:00:00", format="isot", scale="utc")
            )
        }
    )
    af.write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af["obj"].datetime_or_time, Time)


def list_converts_to_list(tmp_path):
    af = asdf.AsdfFile({"obj": UnionObject(collection=[1, 1, 1])})
    af.write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af["obj"].collection, list)


def test_tuple_convert_to_tuple(tmp_path):
    af = asdf.AsdfFile({"obj": UnionObject(collection=(1, 1, 1))})
    af.write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af["obj"].collection, tuple)


@pytest.mark.parametrize(
    "value",
    [
        1,
        2.0,
        "asdf",
        dict(key="value"),
        [1, 2, 3],
        (1, 2, 3),
        {1, 2, 3},
        Time("2023-01-01T00:00:00", format="isot", scale="utc"),
        datetime(2023, 1, 1, 1, 0, 0),
    ],
)
def test_Any_converts_back_to_itself(value, tmp_path):
    af = asdf.AsdfFile({"obj": UnionObject(anything=value)})
    af.write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        if isinstance(value, tuple):
            # If not specified in Pydantic, ASDF will convert tuples to list
            assert isinstance(af["obj"].anything, list)
        else:
            assert isinstance(af["obj"].anything, type(value))



import asdf
import astropy.units as u
from asdf.extension import Extension
from astropy.time import Time
from astropy.units import Quantity

from asdf_pydantic import AsdfPydanticConverter, AsdfPydanticModel


class DataPoint(AsdfPydanticModel):
    _tag = "asdf://asdf-pydantic/examples/tags/datapoint-1.0.0"

    time: Time
    distance: Quantity[u.m]


class DataContainer(AsdfPydanticModel):
    _tag = "asdf://asdf-pydantic/examples/tags/datacontainer-1.0.0"

    positions: list[DataPoint]


def setup_module():
    AsdfPydanticConverter.add_models(DataPoint, DataContainer)

    class TestExtension(Extension):
        extension_uri = "asdf://asdf-pydantic/examples/extensions/test-1.0.0"

        converters = [AsdfPydanticConverter()]  # type: ignore
        tags = [*AsdfPydanticConverter().tags]  # type: ignore

    asdf.get_config().add_extension(TestExtension())


def test_convert_DataPoint_to_asdf(tmp_path):
    af = asdf.AsdfFile(
        {
            "positions": [
                DataPoint(
                    time=Time("2023-01-01T00:00:00"),
                    distance=0 * u.m,
                ),
                DataPoint(
                    time=Time("2023-01-01T01:00:00"),
                    distance=1 * u.m,
                ),
            ]
        }
    ).write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af.tree["positions"], list)
        for position in af.tree["positions"]:
            assert isinstance(position, DataPoint)


def test_convert_DataContainer_to_asdf(tmp_path):
    af = asdf.AsdfFile(
        {
            "container": DataContainer(
                positions=[
                    DataPoint(
                        time=Time("2023-01-01T00:00:00"),
                        distance=0 * u.m,
                    ),
                    DataPoint(
                        time=Time("2023-01-01T01:00:00"),
                        distance=1 * u.m,
                    ),
                ]
            )
        }
    ).write_to(tmp_path / "test.asdf")

    with asdf.open(tmp_path / "test.asdf") as af:
        assert isinstance(af["container"], DataContainer)



from typing import ClassVar

from pydantic import BaseModel


class AsdfPydanticModel(BaseModel):
    """

    ASDF Serialization and Deserialization:
        Serialize to ASDF yaml tree is done with the
        py:classmethod`AsdfPydanticModel.asdf_yaml_tree()` and deserialize to an
        AsdfPydanticModel object with py:meth`AsdfPydanticModel.parse_obj()`.
    """

    _tag: ClassVar[str]

    class Config:
        arbitrary_types_allowed = True

    def asdf_yaml_tree(self) -> dict:
        d = {}
        for field_key, v in self.__dict__.items():
            if field_key not in self.__fields__:
                continue

            if isinstance(v, AsdfPydanticModel):
                d[field_key] = v
            else:
                d[field_key] = self._get_value(
                    v,
                    to_dict=True,
                    by_alias=False,
                    include=None,
                    exclude=None,
                    exclude_unset=False,
                    exclude_defaults=False,
                    exclude_none=False,
                )

        return d



from asdf.extension import Extension

from asdf_pydantic.converter import AsdfPydanticConverter
from asdf_pydantic.examples.shapes import AsdfRectangle
from asdf_pydantic.examples.tree import AsdfNode

AsdfPydanticConverter.add_models(AsdfRectangle, AsdfNode)


class ExampleExtension(Extension):
    extension_uri = "asdf://asdf-pydantic/examples/extensions/examples-1.0.0"
    converters = [AsdfPydanticConverter()]  # type: ignore
    tags = [*AsdfPydanticConverter().tags]  # type: ignore


def get_extensions() -> list[Extension]:
    return [ExampleExtension()]



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel

from asdf_pydantic import AsdfPydanticModel


class Node(BaseModel):
    child: Optional[Node] = None


class AsdfNode(Node, AsdfPydanticModel):
    _tag = "asdf://asdf-pydantic/examples/tags/node-1.0.0"

    child: Optional[Node | AsdfNode] = None



from tempfile import NamedTemporaryFile

import asdf

from asdf_pydantic.examples.extensions import ExampleExtension
from asdf_pydantic.examples.tree import AsdfNode, Node


def setup_module():
    asdf.get_config().add_extension(ExampleExtension())


def test_asdf_node_root_is_AsdfNode():
    node = AsdfNode()
    af = asdf.AsdfFile({"node": node})
    with NamedTemporaryFile() as tempfile:
        af.write_to(tempfile.name)
        with asdf.open(tempfile.name) as ff:
            assert isinstance(ff["node"], AsdfNode)


def test_asdf_node_child_is_AsdfNode():
    node = AsdfNode(child=AsdfNode())
    af = asdf.AsdfFile({"node": node})
    with NamedTemporaryFile() as tempfile:
        af.write_to(tempfile.name)

        with asdf.open(tempfile.name) as ff:
            assert isinstance(ff["node"].child, AsdfNode)


def test_regular_node_child_is_dict():
    node = AsdfNode(child=Node())
    af = asdf.AsdfFile({"node": node})
    with NamedTemporaryFile() as tempfile:
        af.write_to(tempfile.name)

        with asdf.open(tempfile.name) as ff:
            assert isinstance(ff["node"].child, Node)



from __future__ import annotations

from typing import Optional, Type

from asdf.extension import Converter

from asdf_pydantic.model import AsdfPydanticModel

_ASDF_PYDANTIC_SINGLETON_CONVERTER: Optional[AsdfPydanticConverter] = None


class AsdfPydanticConverter(Converter):
    """Implements a converter compatible with all subclass of AsdfPydanticModel.

    The instance is a singleton.
    """

    _tag_to_class: dict[str, Type[AsdfPydanticModel]] = {}

    def __init__(self) -> None:
        global _ASDF_PYDANTIC_SINGLETON_CONVERTER

        if _ASDF_PYDANTIC_SINGLETON_CONVERTER is None:
            _ASDF_PYDANTIC_SINGLETON_CONVERTER = self

        self = _ASDF_PYDANTIC_SINGLETON_CONVERTER

    @classmethod
    def add_models(
        cls, *model_classes: Type[AsdfPydanticModel]
    ) -> "AsdfPydanticConverter":
        for model_class in model_classes:
            cls._tag_to_class[model_class._tag] = model_class
        return cls()

    @property
    def tags(self) -> tuple[str]:
        return tuple(self._tag_to_class.keys())

    @property
    def types(self) -> tuple[str | Type]:
        return tuple(self._tag_to_class.values())

    def select_tag(self, obj, tags, ctx):
        return obj._tag

    def to_yaml_tree(self, obj: AsdfPydanticModel, tag, ctx):
        return obj.asdf_yaml_tree()

    def from_yaml_tree(self, node, tag, ctx):
        return self._tag_to_class[tag].parse_obj(node)



from tempfile import NamedTemporaryFile

import asdf

from asdf_pydantic.examples.shapes import AsdfRectangle


def test_create_asdf_file():
    with NamedTemporaryFile() as tempfile:
        af = asdf.AsdfFile({"rect": AsdfRectangle(width=42, height=10)})
        af.write_to(tempfile.name)



from __future__ import annotations

from asdf_pydantic.examples.tree import AsdfNode, Node


def test_sanity():
    AsdfNode().asdf_yaml_tree() == {"child": None}


def test_should_not_convert_given_child_is_AsdfNode():
    AsdfNode(child=AsdfNode()).asdf_yaml_tree() == {"child": AsdfNode()}


def test_should_convert_given_child_is_Node():
    AsdfNode(child=Node()).asdf_yaml_tree() == {"child": {"child": None}}


def test_given_mix_child_is_mix_of_AsdfNode_and_Node():
    assert AsdfNode(child=AsdfNode(child=Node())).asdf_yaml_tree() == {
        "child": AsdfNode(child=Node())
    }

    assert AsdfNode(child=Node(child=AsdfNode())).asdf_yaml_tree() == {
        "child": {"child": AsdfNode()}
    }
